#
# Core APRS Client
# Various utility routines
# Author: Joerg Schultze-Lutter, 2025
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from expiringdict import ExpiringDict
import datetime
import hashlib
import os
from unidecode import unidecode
import argparse
import zipfile
import apprise
import re
from .client_configuration import program_config
from .client_logger import logger
import sys

# These are global variables which will be used
# in case of an uncaught exception where we send
# the host a final Apprise message along with the
# program's stack trace
exception_occurred = False
ex_type = ex_value = ex_traceback = None

# maximum APRS message lengths with and without trailing
# message number in the outgoing message
#
# When trailing message numbers are activated, the message
# ALWAYS ends with a trailing " (xx/yy)" string whereas
# xx - current message number
# yy - max message number
APRS_MSG_LEN_TRAILING = 59
APRS_MSG_LEN_NOTRAILING = 67


def _get_aprs_msg_len() -> int:
    """
    Internal helper function which determines the maximum length of a single APRS message.
    The corresponding value is dependent on whether the user has activated trailing
    message numbers for the outgoing APRS message or not.

    Parameters
    ==========

    Returns
    =======
    aprs_msg_len: int
        trailing message no enabled: 59
        trailing message no disabled: 67
    """
    if "__main__" not in sys.modules:
        return (
            APRS_MSG_LEN_TRAILING
            if program_config["coac_client_config"]["aprs_message_enumeration"]
            else APRS_MSG_LEN_NOTRAILING
        )
    else:
        return APRS_MSG_LEN_TRAILING


def add_aprs_message_to_cache(
    message_text: str, message_no: str, target_callsign: str, aprs_cache: ExpiringDict
):
    """
    Creates an entry in our expiring dictionary cache. Later on,
    we can check for this entry and see if a certain message has already been sent
    within the past x minutes (setting is specified as part of the definition of the
    ExpiringDict). If we find that entry in our list before that entry has expired,
    we will not send it out again and consider the request to be fulfilled
    Parameters
    ==========
    message_text: str
        APRS message (as extracted from the original incoming message)
    message_no: str
        APRS message number (or 'None' if not present)
    target_callsign: str
        Call sign of the user who has sent this message
    aprs_cache: ExpiringDict
        Reference to the ExpiringDict cache
    Returns
    =======
    aprs_cache: ExpiringDict
        Reference to the ExpiringDict cache, now containing our entry
    """
    # Create message key which consists of:
    # - an md5-ed version of the message text (save some bytes on storage)
    #   Conversion to string is necessary; otherwise, the lookup won't work
    # - the user's call sign
    # - the message number (note that this field's content can be 'None')
    md5_hash = hashlib.md5(message_text.encode("utf-8")).hexdigest()
    key = (md5_hash, target_callsign, message_no)
    # Finally, build the key. Convert it to a tuple as the key needs to be immutable
    key = tuple(key)

    # Add the Key to our expiring cache. The datetime stamp is not used; we
    # just need to give the dictionary entry a value
    aprs_cache[key] = datetime.datetime.now()
    return aprs_cache


def check_if_file_exists(file_name: str):
    """
    Checks if the given file exists. Returns True/False.

    Parameters
    ==========
    file_name: str
                    our file name
    Returns
    =======
    status: bool
        True /False
    """
    return os.path.isfile(file_name)


def get_aprs_message_from_cache(
    message_text: str, message_no: str, target_callsign: str, aprs_cache: ExpiringDict
):
    """
    Checks for an entry in our expiring dictionary cache.
    If we find that entry in our list before that entry has expired,
    we consider the request to be fulfilled and will not process it again
    Parameters
    ==========
    message_text: str
        APRS message (as extracted from the original incoming message)
    message_no: str
        APRS message number (or 'None' if not present)
    target_callsign: str
        Call sign of the user who has sent this message
    aprs_cache: ExpiringDict
        Reference to the ExpiringDict cache
    Returns
    =======
    key: 'Tuple'
        Key tuple (or 'None' if not found / no longer present)
    """
    # Create message key which consists of:
    # - an md5-ed version of the message text (save some bytes on storage)
    #   Conversion to string is necessary; otherwise, the lookup won't work
    # - the user's call sign
    # - the message number (note that this field's content can be 'None')
    md5_hash = hashlib.md5(message_text.encode("utf-8")).hexdigest()
    key = (md5_hash, target_callsign, message_no)
    # Finally, build the key. Convert it to a tuple as the key needs to be immutable
    key = tuple(key)

    if key in aprs_cache:
        return key
    else:
        return None


def dump_string_to_hex(message_text_string: str):
    """
    Converts string to hex format and returns that content to the user.
    If we find that entry in our list before that entry has expired,
    we consider the request to be fulfilled and will not process it again
    Parameters
    ==========
    message_text_string: str
        Text that needs to be converted
    Returns
    =======
    hex-converted text to the user
    """
    return "".join(hex(ord(c))[2:] for c in message_text_string)


def convert_text_to_plain_ascii(message_string: str):
    """
    Converts a string to plain ASCII
    Parameters
    ==========
    message_string: str
        Text that needs to be converted
    Returns
    =======
    hex-converted text to the user
    """
    message_string = (
        message_string.replace("Ä", "Ae")
        .replace("Ö", "Oe")
        .replace("Ü", "Ue")
        .replace("ä", "ae")
        .replace("ö", "oe")
        .replace("ü", "ue")
        .replace("ß", "ss")
    )
    message_string = unidecode(message_string)
    return message_string


def get_command_line_params():
    """
    Gets and returns the command line arguments

    Parameters
    ==========

    Returns
    =======
    configfile: str
        name of the configuration file
    """

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--configfile",
        default="core_aprs_client.cfg",
        type=argparse.FileType("r"),
        help="Program config file name (default is 'core_aprs_client.cfg')",
    )

    args = parser.parse_args()

    configfile = args.configfile.name

    if not check_if_file_exists(file_name=configfile):
        logger.error(msg="Config file does not exist; exiting")
        sys.exit(0)

    return configfile


def build_full_pathname(
    file_name: str,
    root_path_name: str = os.path.abspath(os.getcwd()),
    relative_path_name: str = "data_files",
):
    """
    Build a full-grown path based on $CWD, an optional relative directory name and a file name.

    Parameters
    ==========
    file_name: str
        file name without path
    root_path_name: str
        relative path name that we are going to add.
    relative_path_name: str
        relative path name that we are going to add.

    Returns
    =======
    full_path_name: str
        full path, consisting of root path name, the relative path name and the file name
    """
    return os.path.join(root_path_name, relative_path_name, file_name)


def create_zip_file_from_log(log_file_name: str):
    """
    Creates a zip file from our current log file and
    returns the file name to the caller

    Parameters
    ==========
    log_file_name: str
        our file name, e.g. 'nohup.out'

    Returns
    =======
    success: bool
        True if we were able to create our zip file, otherwise false
    """

    # Check if the file actually exists
    if not log_file_name:
        return False, None
    if not check_if_file_exists(file_name=log_file_name):
        return False, None

    # get a UTC time stamp as reference and create the file name
    _utc = datetime.datetime.now(datetime.UTC)
    zip_file_name = datetime.datetime.strftime(
        _utc, "core_aprs_client_crash_dump_%Y-%m-%d_%H-%M-%S%z.zip"
    )

    # write the zip file to disk
    with zipfile.ZipFile(zip_file_name, mode="w") as archive:
        archive.write(log_file_name)

    # and return the file name
    return True, zip_file_name


def signal_term_handler(signal_number, frame):
    """
    Signal handler for SIGTERM signals. Ensures that the program
    gets terminated in a safe way, thus allowing all databases etc
    to be written to disk.
    Parameters
    ==========
    signal_number:
        The signal number
    frame:
        Signal frame
    Returns
    =======
    """

    logger.debug(msg="Received SIGTERM; forcing clean program exit")
    sys.exit(0)


def generate_apprise_message(
    message_header: str,
    message_body: str,
    apprise_config_file: str,
    message_attachment: str = None,
) -> bool:
    """
    Generates Apprise messages and triggers transmission to the user
    We will use this e.g. for post-mortem dumps in case the client is on the
    verge of crashing

    Parameters
    ==========
    message_header : str
        The message header that we want to send to the user
    message_body : str
        The message body that we want to send to the user
    apprise_config_file: str
        Apprise Yaml configuration file
    message_attachment: str
        The message attachment that we want to send to the user
        'None' if we don't want to send an attachment

    Returns
    =======
    success: bool
        True if successful
    """

    # predefine the output value
    success = False

    logger.debug(msg="Starting Apprise message processing")

    if not apprise_config_file or apprise_config_file == "NOT_CONFIGURED":
        logger.debug(msg="Skipping Apprise messaging; message file is not configured")
        return success

    if not check_if_file_exists(apprise_config_file):
        logger.error(
            msg=f"Apprise config file '{apprise_config_file}' does not exist; aborting"
        )
        return success

    if message_attachment and not check_if_file_exists(message_attachment):
        logger.debug(msg="Attachment file missing; disabling attachments")
        message_attachment = None

    # Create the Apprise instance
    apobj = apprise.Apprise()

    # Create an Config instance
    config = apprise.AppriseConfig()

    # Add a configuration source:
    config.add(apprise_config_file)

    # Make sure to add our config into our apprise object
    apobj.add(config)

    if not message_attachment:
        # Send the notification
        apobj.notify(
            body=message_body,
            title=message_header,
            tag="all",
            notify_type=apprise.NotifyType.FAILURE,
        )
    else:
        # Send the notification
        apobj.notify(
            body=message_body,
            title=message_header,
            tag="all",
            notify_type=apprise.NotifyType.FAILURE,
            attach=message_attachment,
        )

    success = True

    logger.debug(msg="Finished Apprise message processing")
    return success


def check_and_create_data_directory(
    root_path_name: str,
    relative_path_name: str,
):
    """
    Check if the data directory is present and create it, if necessary

    Parameters
    ==========
    root_path_name: str
        relative path name that we are going to add.
    relative_path_name: str
        relative path name that we are going to add.

    Returns
    =======
    success: bool
        False in case of error
    """
    success = True
    _data_directory = os.path.join(root_path_name, relative_path_name)
    if not os.path.exists(_data_directory):
        logger.debug(
            msg=f"Data directory {_data_directory} does not exist, creating ..."
        )
        try:
            os.mkdir(path=_data_directory)
        except OSError:
            logger.error(
                msg=f"Cannot create data directory {_data_directory}, aborting ..."
            )
            success = False
    else:
        if not os.path.isdir(_data_directory):
            logger.error(msg=f"{_data_directory} is not a directory, aborting ...")
            success = False
    return success


def make_pretty_aprs_messages(
    message_to_add: str,
    destination_list: list = None,
    max_len: int = _get_aprs_msg_len(),
    separator_char: str = " ",
    add_sep: bool = True,
    force_outgoing_unicode_messages: bool = False,
) -> list:
    """
    Pretty Printer for APRS messages. As APRS messages are likely to be split
    up (due to the 67 chars message len limitation), this function prevents
    'hard cuts'. Any information that is to be injected into message
    destination list is going to be checked wrt its length. If
    len(current content) + len(message_to_add) exceeds the max_len value,
    the content will not be added to the current list string but to a new
    string in the list.

    Example:

    current APRS message = 1111111111222222222233333333333444444444455555555556666666666

    Add String "Hello World !!!!" (16 chars)

    Add the string the 'conventional' way:

    Message changes to
    Line 1 = 1111111111222222222233333333333444444444455555555556666666666Hello W
    Line 2 = orld !!!!

    This function however returns:
    Line 1 = 1111111111222222222233333333333444444444455555555556666666666
    Line 2 = Hello World !!!!

    In case the to-be-added text exceeds 67 characters due to whatever reason,
    this function first tries to split up the content based on space characters
    in the text and insert the resulting elements word by word, thus preventing
    the program from ripping the content apart. However, if the content consists
    of one or multiple strings which _do_ exceed the maximum text len, then there
    is nothing that we can do. In this case, we will split up the text into 1..n
    chunks of text and add it to the list element.

    Known issues: if the separator_char is different from its default setting
    (space), the second element that is inserted into the list may have an
    additional separator char in the text

    Parameters
    ==========
    message_to_add: str
        message string that is to be added to the list in a pretty way
        If string is longer than 67 chars, we will truncate the information
    destination_list: list
        List with string elements which will be enriched with the
        'mesage_to_add' string. Default: empty list aka user wants new list
    max_len: int:
        Max length of the list's string len.
        The length is dependent on whether the user has activated trailing
        message number information in the outgoing message or not.
        When activated, the message length is 59 - otherwise, it is 67.
        _get_aprs_msg_len() auto-determines the appropriate value.
    separator_char: str
        Separator that is going to be used for dividing the single
        elements that the user is going to add
    add_sep: bool
        True = we will add the separator when more than one item
               is in our string. This is the default
        False = do not add the separator (e.g. if we add the
                very first line of text, then we don't want a
                comma straight after the location
    force_outgoing_unicode_messages: bool
        False = all outgoing UTF-8 content will be down-converted
                to ASCII content
        True = all outgoing UTF-8 content will sent out 'as is'

    Returns
    =======
    destination_list: list
        List array, containing 1..n human readable strings with
        the "message_to_add' input data
    """
    # Dummy handler in case the list is completely empty
    # or a reference to a list item has not been specified at all
    # In this case, create an empty list
    if not destination_list:
        destination_list = []

    # replace non-permitted APRS characters from the
    # message text as APRS-IS might choke on this content
    # Details: see APRS specification pg. 71
    message_to_add = re.sub("[{}|~]+", "", message_to_add)

    # Check if the user wants unicode messages. Default is ASCII
    if (
        not program_config["coac_testing"]["aprsis_enforce_unicode_messages"]
        and not force_outgoing_unicode_messages
    ):
        # Convert the message to plain ascii
        # Unidecode does not take care of German special characters
        # Therefore, we need to 'translate' them first
        message_to_add = convert_text_to_plain_ascii(message_string=message_to_add)

    # If new message is longer than max len then split it up with
    # max chunks of max_len bytes and add it to the array.
    # This should never happen but better safe than sorry.
    # Keep in mind that we only transport plain text anyway.
    if len(message_to_add) > max_len:
        split_data = message_to_add.split()
        for split in split_data:
            # if string is short enough then add it by calling ourself
            # with the smaller text chunk
            if len(split) < max_len:
                destination_list = make_pretty_aprs_messages(
                    message_to_add=split,
                    destination_list=destination_list,
                    max_len=max_len,
                    separator_char=separator_char,
                    add_sep=add_sep,
                    force_outgoing_unicode_messages=force_outgoing_unicode_messages,
                )
            else:
                # string exceeds max len; split it up and add it as is
                string_list = split_string_to_string_list(
                    message_string=split, max_len=max_len
                )
                for msg in string_list:
                    destination_list.append(msg)
    else:  # try to insert
        # Get very last element from list
        if len(destination_list) > 0:
            string_from_list = destination_list[-1]

            # element + new string > max len? no: add to existing string, else create new element in list
            if len(string_from_list) + len(message_to_add) + 1 <= max_len:
                delimiter = ""
                if len(string_from_list) > 0 and add_sep:
                    delimiter = separator_char
                string_from_list = string_from_list + delimiter + message_to_add
                destination_list[-1] = string_from_list
            else:
                destination_list.append(message_to_add)
        else:
            destination_list.append(message_to_add)

    return destination_list


def split_string_to_string_list(
    message_string: str, max_len: int = _get_aprs_msg_len()
):
    """
    Force-split the string into chunks of max_len size and return a list of
    strings. This function is going to be called if the string that the user
    wants to insert exceeds more than e.g. 67 characters. In this unlikely
    case, we may not be able to add the string in a pretty format - but
    we will split it up for the user and ensure that none of the data is lost

    Parameters
    ==========
    message_string: str
        message string that is to be divided into 1..n strings of 'max_len"
        text length
    max_len: int:
        Max length of the list's string len.
        The length is dependent on whether the user has activated trailing
        message number information in the outgoing message or not.
        When activated, the message length is 59 - otherwise, it is 67.
        _get_aprs_msg_len() auto-determines the appropriate value.

    Returns
    =======
    split_strings: list
        List array, containing 1..n strings with a max len of 'max_len'
    """
    split_strings = [
        message_string[index : index + max_len]
        for index in range(0, len(message_string), max_len)
    ]
    return split_strings


def parse_bulletin_data():
    """
    This function parses the bulletin messages from the configuration file,
    checks them for validity and then adds them to the global 'aprs_bulletin_messages'
    dictionary.

    Parameters
    ==========

    Returns
    =======
    """

    # our target directory
    aprs_bulletin_messages: dict = {}

    # Get the key and value from our configuration file's bulletin messages section
    for key, value in program_config["coac_bulletin_messages"].items():
        # Message populated and less than max APRS message length?
        # note: we do not use message enumeration for bulletins
        # therefore, the max length requirement is always fixed (67 bytes)
        if 0 < len(value) <= APRS_MSG_LEN_NOTRAILING:
            # Check if the identifier follows these APRS requirements:
            # 1) must start with fixed "BLN" string
            # 2) needs to be followed by 1..6 ASCII-7 characters and/or digits
            match = re.match(
                pattern=r"^bln[a-z0-9]{1,6}$", string=key, flags=re.IGNORECASE
            )
            if match:
                # We found a match. As a precaution, let's check if the user has
                # used characters in the actual message which are special to APRS
                match = re.findall(r"[{}|~]+", value)
                if match:
                    # Yes, we could remove those characters straight away but the
                    # idea behind this two-fold approach is to inform the user that
                    # ideally, the config file needs to get updated
                    logger.debug(
                        msg=f"APRS bulletin message '{key}': removing special APRS characters from 'value' setting; check your configuration file"
                    )
                    # Let's get rid of those control characters from the message
                    value = re.sub(r"[{}|~]+", "", value)
                # Convert the bulletin identifier to upperkey
                key = key.upper()
                # and add it to our dictionary in case it does not exist
                # and its lenght (after potential character replacements)
                # is still greater than zero
                if key not in aprs_bulletin_messages and len(value) > 0:
                    aprs_bulletin_messages[key] = value
        else:
            # return some clarification to the user on why we have decided to
            # ignore this particular configuration setting
            if len(key) > APRS_MSG_LEN_NOTRAILING:
                logger.debug(
                    f"Ignoring bulletin setting for '{key}'; value exceeds {APRS_MSG_LEN_NOTRAILING} characters. Check your configuration file"
                )
            else:
                logger.debug(f"Ignoring bulletin setting for '{key}'; value is empty")

    return aprs_bulletin_messages


def client_exception_handler():
    """
    This function will be called in case of a regular program exit OR
    an uncaught exception. If an exception has occurred, we will try to
    send out an Apprise message along with the stack trace to the user

    Parameters
    ==========

    Returns
    =======
    """

    if not exception_occurred:
        return

    client_name = program_config["coac_client_config"]["aprs_client_name"]

    # Send a message before we hit the bucket
    message_body = f"'{client_name}' process has crashed. Reason: {ex_value}"

    # Try to zip the log file if possible
    success, log_file_name = create_zip_file_from_log(
        program_config["coac_crash_handler"]["nohup_filename"]
    )

    # check if we can spot a 'nohup' file which already contains our status
    if log_file_name and check_if_file_exists(log_file_name):
        message_body = message_body + " (log file attached)"

    # generate_apprise_message will check again if the file exists or not
    # Therefore, we can skip any further detection steps here
    generate_apprise_message(
        message_header=f"'{client_name}' process has crashed",
        message_body=message_body,
        apprise_config_file=program_config["coac_crash_handler"]["apprise_config_file"],
        message_attachment=log_file_name,
    )


def handle_exception(exc_type, exc_value, exc_traceback):
    """
    Custom exception handler which is installed by the
    main process. We only do a few things:
    - remember that there has been an uncaught exception
    - save the exception type / value / tracebace

    Parameters
    ==========
    exc_type:
        exception type object
    exc_value:
        exception value object
    exc_traceback:
        exception traceback object

    Returns
    =======
    """

    global exception_occurred
    global ex_type
    global ex_value
    global ex_traceback

    # set some global values so that we know why the program has crashed
    exception_occurred = True
    ex_type = exc_type
    ex_value = exc_value
    ex_traceback = exc_traceback

    logger.debug(f"Core process has received uncaught exception: {exc_value}")

    # and continue with the regular flow of things
    sys.__excepthook__(exc_type, exc_value, exc_traceback)


def check_for_default_config():
    """
    Checks if the user tries to run the program with its default configuration
    Doing so does not abort the program but (for now) just generates an error on the command line
    level.

    Parameters
    ==========

    Returns
    =======
    """

    if program_config["coac_client_config"]["aprsis_tocall"] == "APRS":
        logger.error(
            msg="'aprsis_tocall' is still set to default config; change config file ASAP"
        )

    if program_config["coac_client_config"]["aprsis_callsign"] == "COAC":
        logger.error(
            msg="'aprsis_callsign' is still set to default config; change config file ASAP"
        )


def format_list_with_enumeration(mylistarray: list):
    """
    Adds a trailing enumeration to the list if the user has activated this configuration in
    the client's config file

    Parameters
    ==========

    Returns
    =======
    listitem: list
        Either formatted list (if more than one list entry was present) or
        the original list item
    """

    max_total_length = APRS_MSG_LEN_NOTRAILING
    annotation_length = APRS_MSG_LEN_NOTRAILING - APRS_MSG_LEN_TRAILING
    max_content_length = max_total_length - annotation_length

    # check if we have more than 99 entries. We always truncate the list (just to be
    # on the safe side) but whenever more than 99 entries were detected, we also supply
    # the user with a warning message and notify him about the truncation
    if len(mylistarray) > 99:
        logger.warning(
            msg="User has supplied list with more than 99 elements; truncating"
        )
        trimmed_listarray = mylistarray[:98]
        trimmed_listarray.append("[message truncated]")
    else:
        trimmed_listarray = mylistarray
    total = len(trimmed_listarray)

    # now let's add the enumeration to the list - but only if we have
    # more than one list item in our outgoing list
    if len(trimmed_listarray) > 1:
        formatted_list = []
        for i, s in enumerate(trimmed_listarray, start=1):
            annotation = f" ({i:02d}/{total:02d})"
            truncated = s[:max_content_length]
            padded = truncated.ljust(max_content_length)
            final = padded + annotation
            formatted_list.append(final)

        return formatted_list
    else:
        # return the original list to the user
        return trimmed_listarray


def finalize_pretty_aprs_messages(mylistarray: list) -> list:
    """
    Helper method which finalizes the prettified APRS messages
    and triggers the addition of the trailing message numbers (if
    activated by the user).

    Parameters
    ==========
    mylistarray: list
        List of APRS messages

    Returns
    =======
    listitem: list
        Either formatted list (if more than one list entry was present) or
        the original list item
    """
    if program_config["coac_client_config"]["aprs_message_enumeration"]:
        return format_list_with_enumeration(mylistarray=mylistarray)
    else:
        return mylistarray


if __name__ == "__main__":
    pass



#
# Core APRS Client
# Wrapper for APRS-IS communication
# Author: Joerg Schultze-Lutter, 2025
#
# aprslib does not allow us to pass additional parameters to its
# callback function. Therefore, this module acts as a pseudo object in
# order to provide global access to its worker variables
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from .client_logger import logger
import aprslib


class APRSISObject:
    def __init__(
        self, aprsis_callsign, aprsis_passwd, aprsis_host, aprsis_port, aprsis_filter
    ):
        """
        Parameters
        ==========
        aprsis_callsign: str
           Our login callsign
        aprsis_passwd: str
           Our login password
        aprsis_host: str
           Our login hostname
        aprsis_port: int
           Our APS-IS port number
        aprsis_filter: str
           Our APRS-IS filter settings
        """
        self.aprsis_callsign = aprsis_callsign
        self.aprsis_passwd = aprsis_passwd
        self.aprsis_host = aprsis_host
        self.aprsis_port = aprsis_port
        self.aprsis_filter = aprsis_filter
        self.AIS: aprslib.inet.IS = None

        self.ais_open()

    def ais_open(self):
        """
        Helper method for creating the APRS-IS object

        Parameters
        ==========
        self.aprsis_callsign: str
           Our login callsign
        self.aprsis_passwd: str
           Our login password
        self.aprsis_host: str
           Our login hostname
        self.aprsis_port: int
           Our APS-IS port number
        self.aprsis_filter: str
           Our APRS-IS filter settings

        Returns
        =======
        self.AIS: aprslib.inet.IS
           Our APRS-IS object
        """

        logger.debug(
            msg=f"Configuring APRS object: server={self.aprsis_host}, port={self.aprsis_port}, filter={self.aprsis_filter}, APRS-IS passcode={self.aprsis_passwd}, APRS-IS User = {self.aprsis_callsign}"
        )

        self.AIS = aprslib.IS(
            callsign=self.aprsis_callsign,
            passwd=self.aprsis_passwd,
            host=self.aprsis_host,
            port=self.aprsis_port,
        )
        if self.AIS:
            self.AIS.set_filter(self.aprsis_filter)

        return self.AIS

    def ais_start_consumer(self, aprsis_callback: object):
        """
        Helper method for starting the APRS-IS consumer

        Parameters
        ==========
        aprsis_callback: object
           The name of our callback function that we hand over
           to the APRS-IS object

        Returns
        =======
        """
        if type(self.AIS) is aprslib.inet.IS:
            self.AIS.consumer(aprsis_callback, blocking=True, immortal=True, raw=False)
        else:
            logger.debug(msg="Not connected to APRS-IS")

    def ais_connect(self):
        """
        Helper method for connecting to the APRS-IS server

        Parameters
        ==========

        Returns
        =======
        """
        if type(self.AIS) is aprslib.inet.IS:
            self.AIS.connect(blocking=True)

    def ais_is_connected(self):
        """
        Helper method for returning the current connection
        state to the user. Yes, this accesses a protected element
        from the aprslib.inet.IS object, but it still works :-)

        Parameters
        ==========

        Returns
        =======
        """
        if type(self.AIS) is aprslib.inet.IS:
            return self.AIS._connected
        else:
            return False

    def ais_close(self):
        """
        Helper method for closing the APRS-IS connection

        Parameters
        ==========

        Returns
        =======
        """
        # Close APRS-IS connection whereas still present
        if type(self.AIS) is aprslib.inet.IS:
            logger.debug(msg="Closing connection to APRS-IS")
            self.AIS.close()
            self.AIS = None
        else:
            logger.debug(msg="Not connected to APRS-IS")

    def ais_send(self, aprsis_data: str):
        """
        Helper method for starting the APRS-IS consumer

        Parameters
        ==========
        aprsis_data: str
           The data that we want to send to the APRS-IS server

        Returns
        =======
        """
        if type(self.AIS) is aprslib.inet.IS:
            self.AIS.sendall(aprsis_data)
        else:
            logger.debug(msg="Not connected to APRS-IS")

    def ais_get(self):
        """
        Helper method for getting the APRS-IS consumer object

        Parameters
        ==========

        Returns
        =======
        self.AIS: aprslib.inet.IS
            Our APRS-IS object (or 'None')

        """
        if type(self.AIS) is aprslib.inet.IS:
            return self.AIS
        else:
            return None


if __name__ == "__main__":
    pass



#
# Core APRS Client
# Author: Joerg Schultze-Lutter, 2025
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import sys
import signal
import time
import atexit
import os
from functools import partial
import logging
from pprint import pformat
from collections.abc import Callable
from typing import Dict, Any, Mapping
import threading
import copy
from types import MappingProxyType

from . import client_shared
from .client_utils import (
    signal_term_handler,
    check_and_create_data_directory,
    client_exception_handler,
    handle_exception,
    check_for_default_config,
    finalize_pretty_aprs_messages,
    make_pretty_aprs_messages,
    generate_apprise_message,
)
from .client_configuration import load_config, program_config
from .client_aprsobject import APRSISObject
from .client_message_counter import APRSMessageCounter
from .client_expdict import create_expiring_dict
from .client_aprs_communication import (
    aprs_callback,
    init_scheduler_jobs,
    remove_scheduler,
)
from .client_logger import logger, update_logging_level
from .client_return_codes import CoreAprsClientInputParserStatus


class CoreAprsClient:
    config_file: str
    log_level: int
    input_parser: Callable[..., Any]
    output_generator: Callable[..., Any]
    post_processor: Callable[..., Any] | None

    def __init__(
        self,
        config_file: str,
        input_parser: Callable[..., Any],
        output_generator: Callable[..., Any],
        post_processor: Callable[..., Any] | None = None,
        log_level: int = logging.INFO,
    ):
        """
        Class initialization

        Parameters
        ==========
        config_file: str
            'core-aprs-client's config file name, see
            see https://github.com/joergschultzelutter/core-aprs-client/blob/master/docs/configuration.md
        input_parser: Callable[..., Any]
            Name of the user's input parser function, see
            https://github.com/joergschultzelutter/core-aprs-client/blob/master/docs/framework_usage.md
        output_generator: Callable[..., Any]
            Name of the user's output generator function, see
            https://github.com/joergschultzelutter/core-aprs-client/blob/master/docs/framework_usage.md
        post_processor: Callable[..., Any] | None
            (optional); Name of the user's post processor function, see
            https://github.com/joergschultzelutter/core-aprs-client/blob/master/docs/framework_usage.md
        log_level: int
            Log level from Python's 'logging' module
            https://docs.python.org/3/library/logging.html#logging-levels

        Returns
        =======

        """
        self.config_file = config_file
        self.input_parser = input_parser
        self.output_generator = output_generator
        self.post_processor = post_processor
        self.log_level = log_level
        self._dynamic_aprs_bulletins: Dict[str, Any] = {}
        self._lock = threading.Lock()

        # Prepare the config file handler
        # Check if the config file exists
        if not os.path.isfile(config_file):
            raise FileNotFoundError(f"Configuration file '{config_file}' not found")

        # load the program config from our external config file
        load_config(config_file=self.config_file)
        if len(program_config) == 0:
            logger.error(
                msg="Program config file is empty or contains an error; exiting"
            )
            sys.exit(0)

        # And check if the user still runs with the default config
        # Currently, we do not abort the code but only issue an error to the user
        check_for_default_config()

        # Finally, create the MappingProxyType copy of the configuration
        # so that we can expose it to the user, if requested.
        self._config_data = MappingProxyType(program_config.copy())

        # Update the log level (if needed)
        update_logging_level(logging_level=self.log_level)

    def activate_client(self, **kwargs):
        """
        This function is responsible for setting up the communication
        with APRS-IS. It reads the configuration file and establishes
        the network communication with the APRS-IS server. Finally, the
        aprslib's callback function gets triggered.

        Parameters
        ==========
        **kwargs: dict
            Potential user-defined parameters; will get passed along to
            both input parser and output generator

        Returns
        =======

        """

        # Install our custom exception handler, thus allowing us to signal the
        # user who hosts this bot with a message whenever the program is prone to crash
        # OR has ended. In any case, we will then send the file to the host
        #
        # if you are not interested in a post-mortem call stack, remove the following
        # two lines
        logger.debug(msg=f"Activating bot exception handler")
        atexit.register(client_exception_handler)
        sys.excepthook = handle_exception

        # Check whether the data directory exists
        success = check_and_create_data_directory(
            root_path_name=os.path.abspath(os.getcwd()),
            relative_path_name=program_config["coac_data_storage"][
                "aprs_data_directory"
            ],
        )
        if not success:
            sys.exit(0)

        #
        # Read the message counter
        client_shared.aprs_message_counter = APRSMessageCounter(
            file_name=program_config["coac_data_storage"][
                "aprs_message_counter_file_name"
            ]
        )

        # Create the APRS-IS dupe message cache
        client_shared.aprs_message_cache = create_expiring_dict(
            max_len=program_config["coac_dupe_detection"]["msg_cache_max_entries"],
            max_age_seconds=program_config["coac_dupe_detection"][
                "msg_cache_time_to_live"
            ],
        )

        # Register the SIGTERM handler; this will allow a safe shutdown of the program
        logger.debug(msg="Registering SIGTERM handler for safe shutdown...")
        signal.signal(signal.SIGTERM, signal_term_handler)

        # Create the future aprs_scheduler variable
        aprs_scheduler = None

        # Enter the 'eternal' receive loop
        try:
            while True:
                client_shared.AIS = APRSISObject(
                    aprsis_callsign=program_config["coac_client_config"][
                        "aprsis_callsign"
                    ],
                    aprsis_passwd=str(
                        program_config["coac_network_config"]["aprsis_passcode"]
                    ),
                    aprsis_host=program_config["coac_network_config"][
                        "aprsis_server_name"
                    ],
                    aprsis_port=program_config["coac_network_config"][
                        "aprsis_server_port"
                    ],
                    aprsis_filter=program_config["coac_network_config"][
                        "aprsis_server_filter"
                    ],
                )

                # Connect to APRS-IS
                logger.debug(msg="Establishing connection to APRS-IS...")
                client_shared.AIS.ais_connect()

                # Are we connected?
                if client_shared.AIS.ais_is_connected():
                    logger.debug(msg="Established the connection to APRS-IS")

                    # Install the APRS-IS beacon / bulletin schedulers if
                    # activated in the program's configuration file
                    # Otherwise, this field's value will be 'None'
                    aprs_scheduler = init_scheduler_jobs(class_instance=self)

                    # create the partial object for our callback
                    enhanced_callback = partial(
                        aprs_callback,
                        instance=self,
                        parser=self.input_parser,
                        generator=self.output_generator,
                        postproc=self.post_processor,
                        **kwargs,
                    )

                    #
                    # We are now ready to initiate the actual processing
                    # Start the consumer thread
                    logger.info(msg="Starting APRS-IS callback consumer")
                    client_shared.AIS.ais_start_consumer(enhanced_callback)

                    #
                    # We have left the callback, let's clean up a few things
                    logger.debug(msg="Have left the callback consumer")
                    #
                    # First, stop all schedulers. Then remove the associated jobs
                    # This will prevent the beacon/bulletin processes from sending out
                    # messages to APRS_IS
                    # Note that the scheduler might not be active - its existence depends
                    # on the user's configuration file settings.
                    if aprs_scheduler:
                        remove_scheduler(aprs_scheduler=aprs_scheduler)
                    aprs_scheduler = None

                    # close the connection to APRS-IS
                    logger.debug(msg="Closing APRS connection to APRS-IS")
                    client_shared.AIS.ais_close()
                    client_shared.AIS = None
                else:
                    logger.debug(msg="Cannot re-establish connection to APRS-IS")

                # Write current number of packets to disk
                client_shared.aprs_message_counter.write_counter()

                # Enter sleep mode and then restart the loop
                logger.debug(msg=f"Sleeping ...")
                time.sleep(program_config["coac_message_delay"]["packet_delay_message"])

        except (KeyboardInterrupt, SystemExit):
            # Tell the user that we are about to terminate our work
            logger.debug(
                msg="KeyboardInterrupt or SystemExit in progress; shutting down ..."
            )

            # write most recent APRS message counter to disk
            client_shared.aprs_message_counter.write_counter()

            # Shutdown (and remove) the scheduler if it still exists
            if aprs_scheduler:
                remove_scheduler(aprs_scheduler=aprs_scheduler)

            # Close APRS-IS connection whereas still present
            if client_shared.AIS.ais_is_connected():
                client_shared.AIS.ais_close()

    def dryrun_testcall(self, message_text: str, from_callsign: str, **kwargs):
        """
        This function can be used for 100% offline testing. It does trigger
        the standard input parser and output generator.

        Parameters
        ==========
        message_text: str
            The (simulated) APRS input message; sent to us by "from_callsign"
        from_callsign: str
            The callsign that the message was sent from
        **kwargs: dict
            Potential user-defined parameters; will get passed along to
            both input parser and output generator

        Returns
        =======
        none
        """
        logger.info("Activating dryrun testcall...")

        # load the program config from our external config file
        load_config(config_file=self.config_file)
        if len(program_config) == 0:
            logger.error(
                msg="Program config file is empty or contains an error; exiting"
            )
            sys.exit(0)

        # Register the on_exit function to be called on program exit
        atexit.register(client_exception_handler)

        # Set up the exception handler to catch unhandled exceptions
        sys.excepthook = handle_exception

        logger.info(
            msg=f"parsing message '{message_text}' for callsign '{from_callsign}'"
        )

        retcode, input_parser_error_message, response_parameters = self.input_parser(
            instance=self,
            aprs_message=message_text,
            from_callsign=from_callsign,
            **kwargs,
        )

        # this is our potential postprocessor input object
        # If its future value is not 'None' AND a post processor has been
        # set up for the class' object instance, then we try to run the
        # given post processor AFTER the output processor's message has been sent
        # to the user via APRS
        postproc_data = None

        logger.info(msg="Parsed message:")
        logger.info(msg=pformat(response_parameters))
        logger.info(msg=f"return code: {retcode}")
        match retcode:
            case CoreAprsClientInputParserStatus.PARSE_OK:

                logger.info("Running Output Processor build ...")

                # (Try to) build the outgoing message string
                success, output_message_string, postproc_data = self.output_generator(
                    instance=self,
                    input_parser_response_object=response_parameters,
                    **kwargs,
                )
                logger.info(msg=f"Output Processor response={success}, message:")
                logger.info(msg=output_message_string)

                # Generate the outgoing content, if successful
                if success:
                    logger.info(
                        "Output processor status successful; building outgoing messages ..."
                    )
                    # Convert to pretty APRS messaging
                    output_message = make_pretty_aprs_messages(
                        message_to_add=output_message_string
                    )

                    # And finalize the output message, if needed
                    output_message = finalize_pretty_aprs_messages(
                        mylistarray=output_message
                    )
                else:
                    logger.info("Output processor status unsuccessful")
                    # This code branch should never be reached unless there is a
                    # discrepancy between the action determined by the input parser
                    # and the responsive counter-action in the output processor
                    output_message = make_pretty_aprs_messages(
                        message_to_add=program_config["coac_client_config"][
                            "aprs_input_parser_default_error_message"
                        ],
                    )
                logger.info(msg=pformat(output_message))

                if self.post_processor and postproc_data:
                    logger.debug(msg="Activating postprocessor...")
                    success = self.post_processor(
                        instance=self,
                        postprocessor_input_object=postproc_data,
                        **kwargs,
                    )
                    logger.debug(msg=f"postprocessor response='{success}'")

            case CoreAprsClientInputParserStatus.PARSE_ERROR | _:
                logger.error(
                    msg="input_parser_error_message = {input_parser_error_message}"
                )
                # Dump the human readable message to the user if we have one
                if input_parser_error_message:
                    output_message = make_pretty_aprs_messages(
                        message_to_add=f"{input_parser_error_message}",
                    )
                # If not, just dump the link to the instructions
                else:
                    output_message = make_pretty_aprs_messages(
                        message_to_add=program_config["coac_client_config"][
                            "aprs_input_parser_default_error_message"
                        ],
                    )
                # Ultimately, finalize the outgoing message(s) and add the message
                # numbers if the user has requested this in his configuration
                # settings
                output_message = finalize_pretty_aprs_messages(
                    mylistarray=output_message
                )

                logger.info(pformat(output_message))
                logger.info(msg=pformat(response_parameters))

    @property
    def dynamic_aprs_bulletins(self) -> Mapping[str, Any]:
        """
        'getter' for the dynamic aprs bulletins

        Parameters
        ==========

        Returns
        =======
        dynamic_aprs_bulletins: Mapping[str, Any]
            immutable copy of the dynamic aprs bulletins' 'dict' object
        """

        with self._lock:
            return MappingProxyType(self._dynamic_aprs_bulletins)

    @dynamic_aprs_bulletins.setter
    def dynamic_aprs_bulletins(self, new_dict: Dict[str, Any]) -> None:
        """
        'setter' for the dynamic aprs bulletins

        Parameters
        ==========
        new_dict: Mapping[str, Any]
            dynamic aprs bulletins' 'dict' object, used as a foundation
            for the class' dict object

        Returns
        =======

        """
        with self._lock:
            self._dynamic_aprs_bulletins = copy.deepcopy(new_dict)

    @property
    def config_data(self) -> Mapping[str, Any]:
        """
        'getter' for the class' config data

        Parameters
        ==========

        Returns
        =======
        config_data: Mapping[str, Any]
            immutable copy of the class' config data 'dict' object
        """
        with self._lock:
            return self._config_data

    def send_apprise_message(
        self,
        msg_header: str,
        msg_body: str,
        msg_attachment: str | None = None,
        apprise_cfg_file: str | None = None,
    ) -> bool:
        """
        This function is uses the utility function's Apprise messaging
        code in order to send messages via Apprise. For the Apprise config file,
        the user can either specify a separate file OR omits the file name; in that case,
        core-aprs-client will take the Apprise config file name from the framework's config
        file

        Parameters
        ==========
        msg_header: str
        The message header

        msg_body: str
        The message body

        msg_attachment: str|None
        Optional. The message attachment (as a reference to an external file name) or 'None' for no attachment

        apprise_cfg_file: str|None
        Path to an Apprise config file. If set to 'None', we try to use the framework's config file, see
        www.github.com/joergschultzelutter/core-aprs-client/blob/apprise-messaging-method/docs/configuration_subsections/config_crash_handler.md

        Returns
        =======
        success: bool
        The message was sent successfully
        """

        # Check if the user has provided us with an Apprise config file name
        # If not, pick the one that we have in our configuration file
        #
        # Note that there is no need for us to check if the file exists or not: this part
        # will be taken care of by the message generator itself
        if not apprise_cfg_file:
            logger.debug(
                msg="No apprise_cfg_file specified; using default from core-aprs-client's configuration file"
            )
            with self._lock:
                pgm_config = self._config_data
            apprise_cfg_file = pgm_config["coac_crash_handler"]["apprise_config_file"]

        return generate_apprise_message(
            message_header=msg_header,
            message_body=msg_body,
            message_attachment=msg_attachment,
            apprise_config_file=apprise_cfg_file,
        )

    @staticmethod
    def log_info(msg: str) -> None:
        logger.info(msg)

    @staticmethod
    def log_error(msg: str) -> None:
        logger.error(msg)

    @staticmethod
    def log_debug(msg: str) -> None:
        logger.debug(msg)

    @staticmethod
    def log_warning(msg: str) -> None:
        logger.warning(msg)



#
# Core APRS Client
# Wrapper for message counter functions
# Author: Joerg Schultze-Lutter, 2025
#
# aprslib does not allow us to pass additional parameters to its
# callback function. Therefore, this module acts as a pseudo object in
# order to provide global access to its worker variables
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from .client_logger import logger
from .client_utils import build_full_pathname


class APRSMessageCounter:
    def __init__(self, file_name: str):
        """
        This class implements the APRS message counter.

        Parameters
        ==========
        file_name: str
           External file where the APRS message counter is stored.

        Returns
        =======

        """

        # Init our future numeric counter
        self.counter = 0

        # sef the counter's local file name
        self.file_name: str = build_full_pathname(file_name=file_name)

        # and finally, (try to) read the file from disk
        self.read_counter()

    def read_counter(self):
        """
        Reads the latest message counter from a file
        and sets the class' internal counter

        If file is not present, we will start with '0'

        Parameters
        ==========

        Returns
        =======

        """

        logger.debug(msg="Creating APRS message counter object...")

        # Initially, assume that our source file does not exist
        self.counter = 0

        try:
            with open(f"{self.file_name}", "r") as f:
                if f.mode == "r":
                    contents = f.read()
                    f.close()
                    self.counter = int(contents)
        except (FileNotFoundError, Exception):
            self.counter = 0
            logger.debug(
                msg=f"Cannot read content from message counter file {self.file_name}; will create a new file"
            )

    def write_counter(self):
        """
        Writes the latest message counter to a file

        Parameters
        ==========

        Returns
        =======

        """
        logger.debug(msg="Writing APRS message counter object to disk ...")
        try:
            with open(f"{self.file_name}", "w") as f:
                f.write("%d" % self.counter)
                f.close()
        except (IOError, OSError):
            logger.debug(msg=f"Cannot write message counter to {self.file_name}")

    def get_counter(self):
        """
        Getter method for the counter

        Parameters
        ==========

        Returns
        =======
        counter: int
            Our numeric APRS counter

        """
        return self.counter

    def set_counter(self, counter: int):
        """
        Setter method for the counter

        Parameters
        ==========
        counter: int
            Our numeric APRS counter

        Returns
        =======

        """
        if type(counter) is int:
            self.counter = counter
        else:
            raise ValueError("Value type must be int")


if __name__ == "__main__":
    pass



#
# Core APRS Client
# Author: Joerg Schultze-Lutter, 2025
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from enum import Enum

# We support three possible return codes from the input parser:
# PARSE_OK     - Input processor has identified keyword and is ready
#                to continue. This is the desired default state
#                Whenever the return code is PARSE_OK, then we should know
#                by now what the user wants from us. Now, we'll leave it to
#                another module to generate the output data of what we want
#                to send to the user (client_output_generatpr.py).
#                The result to this post-processor will be a general success
#                status code and the message that is to be sent to the user.
# PARSE_ERROR  - an error has occurred. Most likely, the external
#                input processor was either unable to identify a
#                keyword from the message OR a follow-up process has
#                failed; e.g. the user has defined a wx keyword,
#                requiring the sender to supply mandatory location info
#                which was missing from the message. In any way, this signals
#                the callback function that we are unable to process the
#                message any further
# PARSE_IGNORE - The message was ok but we are being told to ignore it. This
#                might be the case if the user's input processor has a dupe
#                check that is additional to the one provided by the
#                core-aprs-client framework. Similar to PARSE_ERROR, we
#                are not permitted to process this request any further BUT
#                instead of sending an error message, we will simply ignore
#                the request. Note that the core-aprs-client framework has
#                already ack'ed the request at this point, thus preventing it
#                from getting resend by APRS-IS over and over again.
#
# Note that you should refrain from using PARSE_IGNORE whenever possible - a
# polite inquiry should always trigger a polite response :-) Nevertheless, there
# might be use cases where you simply need to ignore a (technically valid) request
# in your custom code.


class CoreAprsClientInputParserStatus(Enum):
    PARSE_ERROR = -1
    PARSE_IGNORE = 0
    PARSE_OK = 1


if __name__ == "__main__":
    pass



#
# Core APRS Client
# Wrapper for expiring dictionary / dupe detection
# Author: Joerg Schultze-Lutter, 2025
#
# aprslib does not allow us to pass additional parameters to its
# callback function. Therefore, this module acts as a pseudo object in
# order to provide global access to its worker variables
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from expiringdict import ExpiringDict
from .client_logger import logger


# Helper method for creating our APRS message cache
def create_expiring_dict(max_len: int, max_age_seconds: int):
    """
    Helper method for creating the ExpiringDict

    Parameters
    ==========
    max_len: int
       Number of max dictionary entries
    max_age_seconds: int
       life span per entry in seconds

    Returns
    =======
    """

    # Create the decaying APRS message cache. Any APRS message that is present in
    # this cache will be considered as a duplicate / delayed and will not be processed
    logger.debug(
        msg=f"APRS message dupe cache set to {str(max_len)} max possible entries and a TTL of {str(max_age_seconds / 60)} mins"
    )
    aprs_message_cache = ExpiringDict(
        max_len=max_len,
        max_age_seconds=max_age_seconds,
    )
    return aprs_message_cache


if __name__ == "__main__":
    pass



#
# Core APRS Client
# Various APRS communication routines
# Author: Joerg Schultze-Lutter, 2025
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import time
import types

from .client_configuration import program_config
from .client_utils import (
    make_pretty_aprs_messages,
    get_aprs_message_from_cache,
    add_aprs_message_to_cache,
    parse_bulletin_data,
    finalize_pretty_aprs_messages,
)
from ._version import __version__
from .client_aprsobject import APRSISObject
from . import client_shared
from .client_logger import logger
from .client_return_codes import CoreAprsClientInputParserStatus
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.schedulers import base as apbase
import copy
import re
from collections.abc import Callable
from typing import Any

APRS_MSG_LEN_NOTRAILING = 67


def send_ack(
    myaprsis: APRSISObject,
    target_callsign: str,
    source_msg_no: str,
    source_callsign: str,
    tocall: str,
    packet_delay: float,
    simulate_send: bool = True,
):
    """
    Send acknowledgment for received package to APRS-IS if
    a message number was present
    If 'simulate_send'= True, we still prepare the message but only send it to our log file
    Parameters
    ==========
    myaprsis: aprslib.inet.IS
        Our aprslib object that we will use for the communication part
    target_callsign: str
        Call sign of the user that has sent us the message
    source_msg_no: str
        message number from user's request. Can be 'None'. In that case, we don't send a message acknowledgment to the user
        (normally, we should not enter this function at all if this value is 'None'. The safeguard will still stay in place)
    simulate_send: bool
        If True: Prepare string but only send it to logger
    source_callsign: str
        Our very own APRS callsign (e.g. COAC)
    packet_delay: float
        Delay after sending out our APRS acknowledgment request
    tocall: str
        This bot uses the default TOCALL ("APRS")

    Returns
    =======
    none
    """

    # only prepare an ack if the incoming message contained a message number
    if source_msg_no:
        logger.debug(msg="Preparing acknowledgment receipt")
        # build the ack string
        stringtosend = (
            f"{source_callsign}>{tocall}::{target_callsign:9}:ack{source_msg_no}"
        )
        if not simulate_send:
            logger.debug(msg=f"Sending acknowledgment receipt: {stringtosend}")
            # send the data to APRS-IS
            myaprsis.ais_send(aprsis_data=stringtosend)
            time.sleep(packet_delay)
        else:
            logger.debug(msg=f"Simulating acknowledgment receipt: {stringtosend}")


def send_aprs_message_list(
    myaprsis: APRSISObject,
    message_text_array: list,
    destination_call_sign: str,
    send_with_msg_no: bool,
    aprs_message_counter: int,
    external_message_number: str,
    source_callsign: str,
    tocall: str,
    simulate_send: bool = True,
    new_ackrej_format: bool = False,
    packet_delay: float = 10.0,
    packet_delay_grace_period: float = 1.0,
):
    """
    Send a pre-prepared message list to to APRS_IS
    All packages have a max len of 67 characters
    If 'simulate_send'= True, we still prepare the message but only send it to our log file
    Parameters
    ==========
    myaprsis: aprslib.inet.IS
        Our aprslib object that we will use for the communication part
    message_text_array: list
        Contains 1..n entries of the content that we want to send to the user
    destination_call_sign: str
        Target user call sign that is going to receive the message (usually, this
        is the user's call sign who has sent us the initial message)
    send_with_msg_no: bool
        If True, each outgoing message will have its own message ID attached to the outgoing content
        If False, no message ID is added
    aprs_message_counter: int
        message_counter for messages that require to be ack'ed
    simulate_send: bool
        If True: Prepare string but only send it to logger
    external_message_number: str
        only used if we deal with the new ackrej format
    new_ackrej_format: bool
        false: apply the old ack/rej logic as described in aprs101.pdf.
        We generate our own message id. The user's message ID
        (from the original request) will NOT be added to the
        outgoing message
        ---
        True: apply the new ack/rej logic as described
        in www.aprs.org/aprs11/replyacks.txt
        We generate our own message id. The user's message ID
        (from the original request) WILL be added to the
        outgoing message
    source_callsign: str
        Our very own APRS callsign (e.g. COAC)
    packet_delay: float
        Delay after sending out our APRS acknowledgment request
        Applied in case there are still remaining messages
    packet_delay_grace_period: float
        Delay after sending out our APRS acknowledgment request
        Applied in case there no more still remaining messages
    tocall: str
        This bot uses the default TOCALL ("APRS"). You need to apply
        for your very own TOCALL, see program documentation

    Returns
    =======
    aprs_message_counter: int
        new value for message_counter for messages that require to be ack'ed
    """

    # Send our message list
    for index, single_message in enumerate(message_text_array, start=1):
        # Build the output string
        stringtosend = (
            f"{source_callsign}>{tocall}::{destination_call_sign:9}:{single_message}"
        )
        # Does the outgoing message require to have a message number? (Read:
        # did our INCOMING message request contain a message number, thus requiring
        # us to honor this behavior with our OUTGOING message by adding a message no)?
        if send_with_msg_no:
            # Build the alphanumeric message number
            alpha_counter = get_alphanumeric_counter_value(aprs_message_counter)
            stringtosend = stringtosend + "{" + alpha_counter
            if new_ackrej_format:
                stringtosend = stringtosend + "}" + external_message_number[:2]
            aprs_message_counter = aprs_message_counter + 1
            # reset the message number if we have exceeded the maximum count
            if (
                aprs_message_counter > 676 or alpha_counter == "ZZ"
            ):  # for the alphanumeric counter AA..ZZ, this is equal to "ZZ"
                aprs_message_counter = 0
        # Check if we need to send the message for real or have to simulate it
        if not simulate_send:
            logger.debug(msg=f"Sending response message '{stringtosend}'")
            myaprsis.ais_send(aprsis_data=stringtosend)
        else:
            logger.debug(msg=f"Simulating response message '{stringtosend}'")
        # Apply the regular sleep cycle if there are still messages to be sent
        # Otherwise, use the shorter sleep cycle
        if index < len(message_text_array):
            time.sleep(packet_delay)
        else:
            time.sleep(packet_delay_grace_period)
    return aprs_message_counter


def get_alphanumeric_counter_value(numeric_counter: int):
    """
    Calculate an alphanumeric
    Parameters
    ==========
    numeric_counter: int
        numeric counter that is used for calculating the start value
    Returns
    =======
    alphanumeric_counter: str
        alphanumeric counter that is based on the numeric counter
        Range from AA to ZZ
    """
    first_char = int(numeric_counter / 26)
    second_char = int(numeric_counter % 26)
    alphanumeric_counter = chr(first_char + 65) + chr(second_char + 65)
    return alphanumeric_counter


def send_beacon_and_status_msg(
    class_instance: object,
    myaprsis: APRSISObject,
    aprs_beacon_messages: list,
    simulate_send: bool = True,
):
    """
    Send beacon message list to APRS_IS
    If 'simulate_send'= True, we still prepare the message but only send it to our log file

    Parameters
    ==========
    class_instance: object
        Instance of the main class
    myaprsis: APRSISObject
        Our aprslib object that we will use for the communication part
    aprs_beacon_messages: list
        List of pre-defined APRS beacon messages
    simulate_send: bool
        If True: Prepare string but only send it to logger

    Returns
    =======
    none
    """
    logger.debug(msg="Reached beacon interval; sending beacons")

    # Generate some local variables because the 'black' beautifier seems
    # to choke on multi-dimensional dictionaries
    _aprsis_callsign = program_config["coac_client_config"]["aprsis_callsign"]
    _aprsis_tocall = program_config["coac_client_config"]["aprsis_tocall"]

    # Build and send the list of beacons
    for index, bcn in enumerate(aprs_beacon_messages, start=1):
        # build the beacon string
        stringtosend = f"{_aprsis_callsign}>{_aprsis_tocall}:{bcn}"
        # simulate sending yes/no
        if not simulate_send:
            logger.debug(msg=f"Sending beacon: {stringtosend}")
            # send the beacon to APRS-IS
            myaprsis.ais_send(aprsis_data=stringtosend)
        else:
            logger.debug(msg=f"Simulating beacons: {stringtosend}")
        # apply sleep cycle(s)
        # do we still have messages in our queue?
        # Yes, apply the regular beacon sleep cycle
        if index < len(aprs_beacon_messages):
            time.sleep(program_config["coac_message_delay"]["packet_delay_beacon"])
        else:
            # Otherwise, apply the shorter sleep cycle after sending out
            # our very last beacon message
            time.sleep(
                program_config["coac_message_delay"]["packet_delay_grace_period"]
            )


def send_bulletin_messages(
    class_instance: object,
    myaprsis: APRSISObject,
    bulletin_dict: dict,
    simulate_send: bool = True,
):
    """
    Sends bulletin message list to APRS_IS
    'Recipient' is 'BLNxxx' and is predefined in the bulletin's dict 'key'. The actual message
    itself is stored in the dict's 'value'.
    If 'simulate_send'= True, we still prepare the message but only send it to our log file

    Parameters
    ==========
    class_instance: object
        Instance of the main class
    myaprsis: APRSISObject
        Our aprslib object that we will use for the communication part
    bulletin_dict: dict
        The bulletins that we are going to send upt to the user. Key = BLNxxx, Value = Bulletin Text
    simulate_send: bool
        If True: Prepare string but only send it to logger

    Returns
    =======
    none
    """
    logger.debug(msg="reached bulletin interval; sending bulletins")

    # create our target dictionary which may contain both static
    # and dynamic bulletin messages. First, copy the content from the static
    # bulletin settings - we don't need to check this data again.
    target_dict = copy.deepcopy(bulletin_dict)

    if type(class_instance.dynamic_aprs_bulletins) is types.MappingProxyType:
        # Get the key and value from our configuration file's bulletin messages section
        for key, value in class_instance.dynamic_aprs_bulletins.items():
            # Message populated and less than max APRS message length?
            # note: we do not use message enumeration for bulletins
            # therefore, the max length requirement is always fixed (67 bytes)
            if 0 < len(value) <= APRS_MSG_LEN_NOTRAILING:
                # Check if the identifier follows these APRS requirements:
                # 1) must start with fixed "BLN" string OR "NWS-" string
                # 2) needs to be followed by 1..6 (NWS: 1..5) ASCII-7 characters and/or digits
                # We do not need to look for "finalized" entries, read: the dictionary's key is
                # not required to have keys with 9 chars in total length
                match = re.match(
                    pattern=r"^(bln[a-z0-9]{1,6})|(nws-[a-z0-9]{1,5})$",
                    string=key,
                    flags=re.IGNORECASE,
                )
                if match:
                    # We found a match. As a precaution, let's check if the user has
                    # used characters in the actual message which are special to APRS
                    match = re.findall(r"[{}|~]+", value)
                    if match:
                        logger.debug(
                            msg=f"APRS dynamic bulletin message '{key}': removing special APRS characters from 'value' setting"
                        )
                        # Let's get rid of those control characters from the message
                        value = re.sub(r"[{}|~]+", "", value)
                    # Convert the bulletin identifier to upperkey
                    key = key.upper()
                    # and add it to our dictionary in case it does not exist
                    # and its lenght (after potential character replacements)
                    # is still greater than zero
                    if key not in target_dict and len(value) > 0:
                        target_dict[key] = value
            else:
                logger.debug(
                    f"Ignoring dynamic bulletin setting for '{key}'; value is either empty or exceeds {APRS_MSG_LEN_NOTRAILING} characters. Check your input data"
                )

    # Generate some local variables because the 'black' beautifier seems
    # to choke on multi-dimensional dictionaries
    _aprsis_callsign = program_config["coac_client_config"]["aprsis_callsign"]
    _aprsis_tocall = program_config["coac_client_config"]["aprsis_tocall"]

    # Build and send the list of bulletins
    # Note that we iterate over the unified dictionary of APRS bulletins
    for index, (recipient_id, bln) in enumerate(target_dict.items(), start=1):
        # build the bulletin string
        stringtosend = f"{_aprsis_callsign}>{_aprsis_tocall}::{recipient_id:9}:{bln}"
        # simulate sending yes/no
        if not simulate_send:
            logger.debug(msg=f"Sending bulletin: {stringtosend}")
            # send the bulletin to APRS-IS
            myaprsis.ais_send(aprsis_data=stringtosend)
        else:
            logger.debug(msg=f"simulating bulletins: {stringtosend}")
        # apply sleep cycle(s)
        # do we still have messages in our queue?
        # Yes, apply the regular bulletin sleep cycle
        if index < len(bulletin_dict):
            time.sleep(program_config["coac_message_delay"]["packet_delay_bulletin"])
        else:
            # Otherwise, apply the shorter sleep cycle after sending out
            # our very last bulletin message
            time.sleep(
                program_config["coac_message_delay"]["packet_delay_grace_period"]
            )


# APRSlib callback
# Extract the fields from the APRS message, start the parsing process,
# execute the command and send the command output back to the user
def aprs_callback(
    raw_aprs_packet: dict,
    instance: object,
    parser: Callable[..., Any],
    generator: Callable[..., Any],
    postproc: Callable[..., Any] | None,
    **kwargs,
):
    """
    aprslib callback; this is the core process that takes care of everything
    Parameters
    ==========
    raw_aprs_packet: dict
        dict object, containing the raw APRS data

    parser: Callable[..., Any]
        input parser function
    generator: Callable[..., Any]
        output generator function
    postproc: Callable[..., Any] | None
        optional postprocessing function
    **kwargs: dict
        Potential user-defined parameters; will get passed along to
        both input parser and output generator

    Returns
    =======
    """
    # Get our relevant fields from the APRS message
    addresse_string = raw_aprs_packet.get("addresse")
    message_text_string = raw_aprs_packet.get("message_text")
    response_string = raw_aprs_packet.get("response")
    msgno_string = raw_aprs_packet.get("msgNo")
    from_callsign = raw_aprs_packet.get("from")
    format_string = raw_aprs_packet.get("format")
    ack_msgno_string = raw_aprs_packet.get("ackMsgNo")

    # lower the response in case we received one
    if response_string:
        response_string = response_string.lower()

    # Check if we need to deal with the old vs the new message format
    new_ackrej_format = True if ack_msgno_string else False

    # Check if this request supports a msgno
    msg_no_supported = True if msgno_string else False

    # User's call sign. read: who has sent us this message?
    if from_callsign:
        from_callsign = from_callsign.upper()

    if addresse_string:
        # Lets examine what we've got:
        # 1. Message format should always be 'message'.
        #    This is even valid for ack/rej responses
        # 2. Message text should contain content
        # 3. response text should NOT be ack/rej
        # Continue if both assumptions are correct
        if (
            format_string == "message"
            and message_text_string
            and response_string not in ["ack", "rej"]
        ):
            # This is a message that belongs to us
            #
            # Check if the message is present in our decaying message cache
            # If the message can be located, then we can assume that we have
            # processed (and potentially acknowledged) that message request
            # within the last e.g. 5 minutes and that this is a delayed / dupe
            # request, thus allowing us to ignore this request.
            aprs_message_key = get_aprs_message_from_cache(
                message_text=message_text_string,
                message_no=msgno_string,
                target_callsign=from_callsign,
                aprs_cache=client_shared.aprs_message_cache,
            )
            if aprs_message_key:
                logger.debug(
                    msg="DUPLICATE APRS PACKET - this message is still in our decaying message cache"
                )
                logger.debug(
                    msg=f"Ignoring duplicate APRS packet raw_aprs_packet: {raw_aprs_packet}"
                )
            else:
                logger.debug(msg=f"Received raw_aprs_packet: {raw_aprs_packet}")

                # Send an ack if we DID receive a message number
                # and we DID NOT have received a request in the
                # new ack/rej format
                # see aprs101.pdf pg. 71ff.
                if msg_no_supported and not new_ackrej_format:
                    send_ack(
                        myaprsis=client_shared.AIS,
                        simulate_send=program_config["coac_testing"][
                            "aprsis_simulate_send"
                        ],
                        source_callsign=program_config["coac_client_config"][
                            "aprsis_callsign"
                        ],
                        tocall=program_config["coac_client_config"]["aprsis_tocall"],
                        target_callsign=from_callsign,
                        source_msg_no=msgno_string,
                        packet_delay=program_config["coac_message_delay"][
                            "packet_delay_ack"
                        ],
                    )
                #
                # This is where the magic happens: Try to figure out what the user
                # wants from us. If we were able to understand the user's message,
                # 'success' will be true. In any case, the 'response_parameters'
                # dictionary will give us a hint about what to do next (and even
                # contains the parser's error message if 'success' != True)
                # input parameters: the actual message, the user's call sign and
                # the aprs.fi API access key for location lookups
                #
                # Note: we call the function which was passed along with the
                # callback object
                retcode, input_parser_error_message, response_parameters = parser(
                    instance=instance,
                    aprs_message=message_text_string,
                    from_callsign=from_callsign,
                    **kwargs,
                )
                logger.debug(msg=f"Input parser result: {retcode}")
                logger.debug(msg=response_parameters)

                # this is our future output message object
                output_message = []

                # this is our potential postprocessor input object
                # If its future value is not 'None' AND a post processor has been
                # set up for the class' object instance, then we try to run the
                # given post processor AFTER the output processor's message has been sent
                # to the user via APRS
                postproc_data = None

                #
                # parsing successful?
                #
                # We support three possible return codes from the input parser:
                # PARSE_OK     - Input processor has identified keyword and is ready
                #                to continue. This is the desired default state
                #                Whenever the return code is PARSE_OK, then we should know
                #                by now what the user wants from us. Now, we'll leave it to
                #                another module to generate the output data of what we want
                #                to send to the user (client_output_generatpr.py).
                #                The result to this post-processor will be a general success
                #                status code and the message that is to be sent to the user.
                # PARSE_ERROR  - an error has occurred. Most likely, the external
                #                input processor was either unable to identify a
                #                keyword from the message OR a follow-up process has
                #                failed; e.g. the user has defined a wx keyword,
                #                requiring the sender to supply mandatory location info
                #                which was missing from the message. In any way, this signals
                #                the callback function that we are unable to process the
                #                message any further
                # PARSE_IGNORE - The message was ok but we are being told to ignore it. This
                #                might be the case if the user's input processor has a dupe
                #                check that is additional to the one provided by the
                #                core-aprs-client framework. Similar to PARSE_ERROR, we
                #                are not permitted to process this request any further BUT
                #                instead of sending an error message, we will simply ignore
                #                the request. Note that the core-aprs-client framework has
                #                already ack'ed the request at this point, thus preventing it
                #                from getting resend by APRS-IS over and over again.
                #
                # Note that you should refrain from using PARSE_IGNORE whenever possible - a
                # polite inquiry should always trigger a polite response :-) Nevertheless, there
                # might be use cases where you simply need to ignore a (technically valid) request
                # in your custom code.
                #
                #
                match retcode:
                    case CoreAprsClientInputParserStatus.PARSE_OK:
                        # Generate the output message for the requested keyword
                        #
                        # Note: we call the function which was passed along with the
                        # callback object
                        success, output_string, postproc_data = generator(
                            instance=instance,
                            input_parser_response_object=response_parameters,
                            **kwargs,
                        )
                        if success:
                            output_message = make_pretty_aprs_messages(
                                message_to_add=output_string
                            )
                        else:
                            # This code branch should never be reached unless there is a
                            # discrepancy between the action determined by the input parser
                            # and the responsive counter-action from the output processor
                            output_message = make_pretty_aprs_messages(
                                message_to_add=program_config["coac_client_config"][
                                    "aprs_input_parser_default_error_message"
                                ],
                            )
                    # This is the branch where the input parser failed to understand
                    # the message. A possible reason: you sent a keyword which requires
                    # an additional parameter but failed to send that one, too.
                    # As we only parse but never process data in that input
                    # parser, we sinply don't know what to do with the user's message
                    # and get back to him with a generic response.
                    case CoreAprsClientInputParserStatus.PARSE_ERROR:
                        # Dump the human readable message to the user if we have one
                        if input_parser_error_message:
                            output_message = make_pretty_aprs_messages(
                                message_to_add=f"{input_parser_error_message}",
                            )
                        # If not, just dump the link to the instructions
                        # This is the default branch which dumps generic information
                        # to the client whenever there is no generic error text from the input parser
                        else:
                            output_message = make_pretty_aprs_messages(
                                message_to_add=program_config["coac_client_config"][
                                    "aprs_input_parser_default_error_message"
                                ],
                            )
                            logger.debug(
                                msg=f"Unable to process APRS packet {raw_aprs_packet}"
                            )
                    # default branch for anything else, including PARSE_IGNORE
                    case _:
                        pass

                # Ultimately, finalize the outgoing message(s) and add the message
                # numbers if the user has requested this in his configuration
                # settings
                output_message = finalize_pretty_aprs_messages(
                    mylistarray=output_message
                )

                # Send our message(s) to APRS-IS
                _aprs_msg_count = send_aprs_message_list(
                    myaprsis=client_shared.AIS,
                    simulate_send=program_config["coac_testing"][
                        "aprsis_simulate_send"
                    ],
                    message_text_array=output_message,
                    destination_call_sign=from_callsign,
                    send_with_msg_no=msg_no_supported,
                    aprs_message_counter=client_shared.aprs_message_counter.get_counter(),
                    external_message_number=msgno_string,
                    new_ackrej_format=new_ackrej_format,
                    source_callsign=program_config["coac_client_config"]["aprsis_callsign"],
                    tocall=program_config["coac_client_config"]["aprsis_tocall"],
                    packet_delay=program_config["coac_message_delay"][
                        "packet_delay_message"
                    ],
                    packet_delay_grace_period=program_config["coac_message_delay"][
                        "packet_delay_grace_period"
                    ],
                )

                # And store the new APRS message number in our counter object
                client_shared.aprs_message_counter.set_counter(_aprs_msg_count)

                # We've finished processing this message. Update the decaying
                # cache with our message.
                # Store the core message data in our decaying APRS message cache
                # Dupe detection is applied regardless of the message's
                # processing status
                client_shared.aprs_message_cache = add_aprs_message_to_cache(
                    message_text=message_text_string,
                    message_no=msgno_string,
                    target_callsign=from_callsign,
                    aprs_cache=client_shared.aprs_message_cache,
                )

                # Finally, execute the post processor function but ONLY if the user has
                # forwarded a function to us AND we have received some postprocessor-specific
                # input from the output generator function - which indicates to us that the
                # user actually wants us to that postprocessor step
                if postproc_data and postproc:
                    _ = postproc(
                        instance=instance,
                        postprocessor_input_object=postproc_data,
                        **kwargs,
                    )


def init_scheduler_jobs(class_instance: object):
    """
    Initializes the scheduler jobs for APRS bulletins and / or beacons.

    Parameters
    ==========

    Returns
    =======
    my_scheduler: BackgroundScheduler object or 'None' if no scheduler was initialized.
    """

    if (
        program_config["coac_beacon_config"]["aprsis_broadcast_beacon"]
        or program_config["coac_bulletin_config"]["aprsis_broadcast_bulletins"]
    ):
        # If we reach this position in the code, we have at least one
        # task that needs to be scheduled (bulletins and/or position messages
        #
        # Create the scheduler
        my_scheduler = BackgroundScheduler()

        # Install two schedulers tasks, if requested by the user
        # The first task is responsible for sending out beacon messages
        # to APRS; it will be triggered every 30 mins
        #

        # The 2nd task is responsible for sending out bulletin messages
        # to APRS; it will be triggered every 4 hours
        #

        if program_config["coac_beacon_config"]["aprsis_broadcast_beacon"]:
            # Send initial beacon after establishing the connection to APRS_IS
            logger.debug(
                msg="Send initial beacon after establishing the connection to APRS_IS"
            )

            #
            # APRS_IS beacon messages (will be sent every 30 mins)
            # - APRS Position (first line) needs to have 63 characters or less
            # - APRS Status can have 67 chars (as usual)
            # Details: see aprs101.pdf chapter 8
            #
            # The client will NOT check the content and send it out 'as is'
            #
            # This message is a position report; format description can be found on pg. 23ff and pg. 94ff.
            # of aprs101.pdf. Message symbols: see http://www.aprs.org/symbols/symbolsX.txt and aprs101.pdf
            # on page 104ff.
            # Format is as follows: =Lat primary-symbol-table-identifier lon symbol-identifier test-message
            # Lat/lon from the configuration have to be valid or the message will not be accepted by aprs-is
            #
            # Example nessage: COAC>APRS:=5150.34N/00819.60E?COAC 0.01
            # results in
            # lat = 5150.34N
            # primary symbol identifier = /
            # lon = 00819.60E
            # symbol identifier = ?
            # plus some text.
            # The overall total symbol code /? refers to a server icon - see list of symbols
            #
            # as all of our parameters are stored in a dictionary, we need to construct

            # create a couple of local variables as the 'black' prettifier seems to
            # choke on multi-dimensional dictionaries

            # fmt:off
            _aprsis_latitude = program_config["coac_beacon_config"]["aprsis_latitude"]
            _aprsis_longitude = program_config["coac_beacon_config"]["aprsis_longitude"]
            _aprsis_table = program_config["coac_beacon_config"]["aprsis_table"]
            _aprsis_symbol = program_config["coac_beacon_config"]["aprsis_symbol"]
            _aprsis_callsign = program_config["coac_client_config"]["aprsis_callsign"]
            _aprsis_beacon_altitude_ft = program_config["coac_beacon_config"]["aprsis_beacon_altitude_ft"]
            # fmt:on

            # check if altitude data is present
            #
            # default assumption: User has provided a numeric value which is then
            # transposed to altitude settings; the config file importer has already
            # done this job for us by now. However, if the variable's type is "str",
            # there is a chance that the user did not provide us with altitude information
            # so we need to check on whether we have to include this field in our
            # beacon or not
            #
            # Assume the default case first (user HAS provided altitude)
            _altitude_present = True
            # field type = 'str'? Then check if the field is empty
            if type(_aprsis_beacon_altitude_ft) is str:
                # Note that this should ALWAYS return "false" - otherwise,
                # we might have encountered a cfg value in the config file
                # which did not consist of purely numeric content
                _altitude_present = len(_aprsis_beacon_altitude_ft) > 0
                if _altitude_present:
                    if not bool(re.fullmatch(r"\d+", _aprsis_beacon_altitude_ft)):
                        raise ValueError(
                            "Invalid 'altitude' setting in configuration file; value must either be numeric or empty"
                        )

            # Build the proper altitude string if altitude data is present
            _aprsis_beacon_altitude_ft = (
                f" /A={str(_aprsis_beacon_altitude_ft).zfill(6)[:6]}"
                if _altitude_present
                else ""
            )

            # generate the APRS beacon string
            _beacon = f"={_aprsis_latitude}{_aprsis_table}{_aprsis_longitude}{_aprsis_symbol}{_aprsis_callsign} {__version__}{_aprsis_beacon_altitude_ft}"

            # and store it in a list item
            aprs_beacon_messages: list = [_beacon]

            # Ultimately, send the initial beacon message
            send_beacon_and_status_msg(
                class_instance=class_instance,
                myaprsis=client_shared.AIS,
                aprs_beacon_messages=aprs_beacon_messages,
                simulate_send=program_config["coac_testing"]["aprsis_simulate_send"],
            )

            # Now let's add position beaconing to scheduler
            my_scheduler.add_job(
                send_beacon_and_status_msg,
                "interval",
                id="aprsbeacon",
                minutes=program_config["coac_beacon_config"][
                    "aprsis_beacon_interval_minutes"
                ],
                args=[
                    class_instance,
                    client_shared.AIS,
                    aprs_beacon_messages,
                    program_config["coac_testing"]["aprsis_simulate_send"],
                ],
                max_instances=1,
                coalesce=True,
            )

        if program_config["coac_bulletin_config"]["aprsis_broadcast_bulletins"]:
            # prepare the bulletin data
            aprs_bulletin_messages = parse_bulletin_data()

            # Install scheduler task 2 - send standard bulletins (advertising the program instance)
            # The bulletin messages consist of fixed content and are defined at the beginning of
            # this program code
            my_scheduler.add_job(
                send_bulletin_messages,
                "interval",
                id="aprsbulletin",
                minutes=program_config["coac_bulletin_config"][
                    "aprsis_bulletin_interval_minutes"
                ],
                args=[
                    class_instance,
                    client_shared.AIS,
                    aprs_bulletin_messages,
                    program_config["coac_testing"]["aprsis_simulate_send"],
                ],
                max_instances=1,
                coalesce=True,
            )

        # Ultimately, start the scheduler
        my_scheduler.start()
    # Default handler in case we neither want bulletins nor beacons
    else:
        my_scheduler = None

    return my_scheduler


def remove_scheduler(aprs_scheduler: BackgroundScheduler):
    """
    Shuts down and the scheduler whereas present.

    Parameters
    ==========
    aprs_scheduler: BackgroundScheduler object or None if no scheduler was initialized.

    Returns
    =======

    """
    # If the scheduler object exists, then try to pause it before it gets destroyed
    if type(aprs_scheduler) is BackgroundScheduler:
        logger.debug(msg="Pausing aprs_scheduler")
        aprs_scheduler.pause()
        aprs_scheduler.remove_all_jobs()
        logger.debug(msg="Shutting down aprs_scheduler")
        if aprs_scheduler.state != apbase.STATE_STOPPED:
            try:
                aprs_scheduler.shutdown()
            except:
                logger.debug(msg="Exception during scheduler shutdown SystemExit loop")


if __name__ == "__main__":
    pass



#
# Core APRS Client
# Shared variables for 'aprslib' callback routine
# Author: Joerg Schultze-Lutter, 2025
#
# aprslib does not allow us to pass additional parameters to its
# callback function. Therefore, this module acts as a pseudo object in
# order to provide global access to its worker variables
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

AIS = None
aprs_message_counter = None
aprs_message_cache = None

if __name__ == "__main__":
    pass



#
# Core APRS Client
# Wrapper for logger object
# Author: Joerg Schultze-Lutter, 2025
#
# aprslib does not allow us to pass additional parameters to its
# callback function. Therefore, this module acts as a pseudo object in
# order to provide global access to its worker variables
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import logging

# Change the logging level to logging.DEBUG if you want to see more
# content in the program's log file.
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(module)s -%(levelname)s - %(message)s",
)
logger = logging.getLogger(__name__)


def update_logging_level(logging_level: int):
    logger.setLevel(logging_level)
    for handler in logger.handlers:
        handler.setLevel(logging_level)


if __name__ == "__main__":
    pass



#!/usr/bin/env/python
#
# Setup file for PyPi publishing work flow via GitHub Actions
# Author: Joerg Schultze-Lutter, 2025
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from setuptools import setup, find_packages
import os
import re
from pathlib import Path
from urllib.parse import urljoin, urlsplit, urlunsplit


# Amend this section with your custom data
PACKAGE_SOURCE_DIR = "src"
PACKAGE_NAME = "core-aprs-client"
DESCRIPTION = "Extensible APRS bot framework with dupe detection, beacon/bulletin support and other nice features. Just add your custom APRS bot functions - the APRS bot framework will take care of the rest."
AUTHOR = "Joerg Schultze-Lutter"
AUTHOR_EMAIL = "joerg.schultze.lutter@gmail.com"
URL = "https://github.com/joergschultzelutter/core-aprs-client"
# https://pypi.org/classifiers/
CLASSIFIERS = [
    "Intended Audience :: Developers",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Topic :: Software Development",
    "Operating System :: OS Independent",
    "Development Status :: 5 - Production/Stable",
]
INSTALL_REQUIRES = [
    "aprslib>=0.7.2",
    "apprise>=1.9.4",
    "expiringdict>=1.2.2",
    "unidecode>=1.4.0",
    "apscheduler>=3.11.0",
]
KEYWORDS = ["Ham Radio", "Amateur Radio", "APRS"]
LICENSE = "GNU General Public License v3 (GPLv3)"


# check if the workflow is triggered from within a GitHub action or executed as standalone code
def running_in_a_github_action():
    return os.getenv("GITHUB_ACTIONS") == "true"


# Absolute URL to base repository
BASE_URL = "https://github.com/joergschultzelutter/core-aprs-client/blob/master/"

# Markdown link recognition (will not work on image links)
LINK_RE = re.compile(r"(?<!\!)\[(?P<text>[^\]]+)\]\((?P<href>[^)\s]+)\)")


def is_absolute_url(href: str) -> bool:
    return href.startswith(("http://", "https://", "mailto:"))


def is_root_relative(href: str) -> bool:
    return href.startswith("/")


def looks_like_md(href: str) -> bool:
    path = urlsplit(href).path
    return path.lower().endswith(".md")


def to_absolute_md(href: str) -> str:
    parts = urlsplit(href)  # (scheme, netloc, path, query, fragment)
    abs_url = urljoin(BASE_URL, parts.path)
    return urlunsplit(
        urlsplit(abs_url)._replace(query=parts.query, fragment=parts.fragment)
    )


def rewrite_md_links(md_text: str) -> str:
    def repl(m: re.Match) -> str:
        text = m.group("text")
        href = m.group("href")

        if (
            not is_absolute_url(href)
            and not is_root_relative(href)
            and looks_like_md(href)
        ):
            href = to_absolute_md(href)

        return f"[{text}]({href})"

    return LINK_RE.sub(repl, md_text)


def transform_file(filename: str) -> str:
    in_path = Path(filename)
    if not in_path.exists():
        raise FileNotFoundError(filename)

    content = in_path.read_text(encoding="utf-8")
    transformed = rewrite_md_links(content)
    return transformed


if __name__ == "__main__":
    LONG_DESCRIPTION = transform_file("README.md")

    # PLACEHOLDER for manual package installations via local requirements file
    # git+https://github.com/joergschultzelutter/core-aprs-client#egg=core-aprs-client
    GITHUB_PROGRAM_VERSION = "0.0.1"

    if running_in_a_github_action():
        # Only run this branch if we are part of a GitHub action. Otherwise, just skip
        # it as we won't be able to get the version info from GitHub anyway
        #
        # get VERSION value from GitHub workflow and terminate workflow if value is None
        GITHUB_PROGRAM_VERSION = os.getenv("GITHUB_PROGRAM_VERSION")
        if not GITHUB_PROGRAM_VERSION:
            raise ValueError("Did not receive release label version info from GitHub")
    else:
        if len(GITHUB_PROGRAM_VERSION) == 0:
            raise ValueError(
                "Manual run requires a manually set GITHUB_PROGRAM_VERSION; change setup.py accordingly"
            )

    # Main setup branch
    setup(
        name=PACKAGE_NAME,
        version=GITHUB_PROGRAM_VERSION,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        url=URL,
        packages=find_packages(where=PACKAGE_SOURCE_DIR),
        package_dir={"": PACKAGE_SOURCE_DIR},
        include_package_data=True,
        classifiers=CLASSIFIERS,
        license=LICENSE,
        install_requires=INSTALL_REQUIRES,
        keywords=KEYWORDS,
    )



#
# Core APRS Client
# Schema file for the configuration file parser
# Author: Joerg Schultze-Lutter, 2025
#
# Used for a validation of the configuration file data. Every configuration
# file section that starts with a leading "coac_" AND is NOT part of this
# file's EXCLUDED_CONFIGURATION_SCHEMA section will get validated against:
#
# - expected configuration file sections
# - within these sections:
#   - expected variable names
#   - expected variable types
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# This section defines the configuration data that we _want_ to validate
# Literally speaking, all configuration sections starting with a leading
# "coac_" string are validated
CONFIGURATION_SCHEMA = {
    "coac_client_config": {
        "aprsis_callsign": str,
        "aprsis_tocall": str,
        "aprs_client_name": str,
        "aprs_input_parser_default_error_message": str,
        "aprs_message_enumeration": bool,
    },
    "coac_network_config": {
        "aprsis_server_name": str,
        "aprsis_server_port": int,
        "aprsis_passcode": int,
        "aprsis_server_filter": str,
    },
    "coac_beacon_config": {
        "aprsis_broadcast_beacon": bool,
        "aprsis_table": str,
        "aprsis_symbol": str,
        "aprsis_latitude": str,
        "aprsis_longitude": str,
        "aprsis_beacon_altitude_ft": int,
        "aprsis_beacon_interval_minutes": int,
    },
    "coac_bulletin_config": {
        "aprsis_broadcast_bulletins": bool,
        "aprsis_bulletin_interval_minutes": int,
    },
    "coac_crash_handler": {
        "apprise_config_file": str,
        "nohup_filename": str,
    },
    "coac_dupe_detection": {
        "msg_cache_max_entries": int,
        "msg_cache_time_to_live": int,
    },
    "coac_message_delay": {
        "packet_delay_message": float,
        "packet_delay_ack": float,
        "packet_delay_grace_period": float,
        "packet_delay_bulletin": float,
        "packet_delay_beacon": float,
    },
    "coac_testing": {
        "aprsis_enforce_unicode_messages": bool,
        "aprsis_simulate_send": bool,
    },
    "coac_data_storage": {
        "aprs_data_directory": str,
        "aprs_message_counter_file_name": str,
    },
}

# This section defines the configuration data that we want to
# _exclude_ from validation. We cannot validate the bulletin
# messages as they represent dynamic data which can be changed
# by the end user
EXCLUDED_CONFIGURATION_SCHEMA = ["coac_bulletin_messages"]


if __name__ == "__main__":
    pass



#
# Core APRS Client
# Configuration file import and validation
# Imports the configuration data into a dictionary object
# and performs a generic validation against predefined schema data
# Author: Joerg Schultze-Lutter, 2025
#
# aprslib does not allow us to pass additional parameters to its
# callback function. Therefore, this module acts as a pseudo object in
# order to encapsulate the client configuration data
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
import configparser
from os import path
from .client_configuration_schema import (
    CONFIGURATION_SCHEMA,
    EXCLUDED_CONFIGURATION_SCHEMA,
)

config = configparser.ConfigParser()
program_config = {}


def load_config(config_file: str):
    """
    Loads the program configuration from the config file
    Returns the program configuration as a dictionary
    If the config file does not exist, the dictionary will be empty

    Parameters
    ==========
    config_file: 'str'
        Name of the configuration file

    Returns
    =======
    none
    """

    if path.isfile(config_file):
        try:
            config.read(config_file)
            config_to_dict(config)
        except:
            program_config.clear()
    else:
        program_config.clear()
    validate_config_schema(program_config)


def _parse_value(value: str):
    """
    Helper method to convert a string to its native value format

    Parameters
    ==========
    value: 'str'
       Our input value

    Returns
    =======
    Converted data type
    """
    if value.lower() in {"true", "yes", "on"}:
        return True
    elif value.lower() in {"false", "no", "off"}:
        return False
    try:
        if "." in value:
            return float(value)
        return int(value)
    except ValueError:
        return value


def config_to_dict(myconfig: configparser.ConfigParser):
    """
    Converts a ConfigParser object to a dictionary

    Parameters
    ==========
    myconfig: configparser.ConfigParser
       ConfigParser input object

    Returns
    =======
    program_config: dict
        Our program configuration dictionary
    """
    program_config.clear()
    for section in myconfig.sections():
        program_config[section] = {
            key: _parse_value(value) for key, value in config.items(section)
        }
    return program_config


def get_config():
    """
    Helper method: gets the program configuration dictionary

    Parameters
    ==========

    Returns
    =======
    program_config: 'dict'
        Our program configuration dictionary
    """
    return program_config


def validate_config_schema(cfg: dict):
    """
    Helper method: validates config file data against
    predefined schema and checks for missing fields and/or
    fields with invalid data types

    Parameters
    ==========
    cfg: dict
        Dictionary with data from config file

    Returns
    =======
    """
    for section, values in cfg.items():
        if not section.startswith("coac_"):
            continue

        expected_schema = CONFIGURATION_SCHEMA.get(section)
        if not expected_schema:
            if section in EXCLUDED_CONFIGURATION_SCHEMA:
                continue
            else:
                raise KeyError(
                    f"Schema definition for section '{section}' is missing from the configuration file"
                )

        # a) Check all required keys are present
        missing_keys = set(expected_schema.keys()) - set(values.keys())
        if missing_keys:
            raise KeyError(
                f"Configuration file section '{section}': missing keys {missing_keys}"
            )

        # b) Check type correctness
        for key, expected_type in expected_schema.items():
            if key not in values:
                continue
            actual_value = values[key]
            if not isinstance(actual_value, expected_type):
                raise TypeError(
                    f"Configuration file section '{section}': key '{key}' has wrong type "
                    f"(expected {expected_type.__name__}, got {type(actual_value).__name__})"
                )


if __name__ == "__main__":
    pass
